@@ -1,11 +1,32 @@
 # -*- coding: utf-8 -*-
+
+# Copyright © 2012-2016 marmuta <marmvta@gmail.com>
+#
+# This file is part of Onboard.
+#
+# Onboard is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 3 of the License, or
+# (at your option) any later version.
+#
+# Onboard is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program. If not, see <http://www.gnu.org/licenses/>.
+
 """ GTK specific keyboard class """
 
 from __future__ import division, print_function, unicode_literals
 
+import time
 import logging
 _logger = logging.getLogger(__name__)
 
+from Onboard.Version import require_gi_versions
+require_gi_versions()
 try:
     from gi.repository import Atspi
 except ImportError as e:
@@ -17,6 +38,7 @@ from Onboard.utils        import Rect, EventSource, Process, unicode_str
 from Onboard.Config import Config
 config = Config()
 
+
 class AsyncEvent:
     """
     Decouple AT-SPI events from D-Bus calls to reduce the risk for deadlocks.
@@ -27,9 +49,9 @@ class AsyncEvent:
 
     def __repr__(self):
         return type(self).__name__ + "(" + \
-           ", ".join(str(key) + "=" + repr(val) \
-                     for key, val in self._kwargs.items()) \
-           + ")"
+            ", ".join(str(key) + "=" + repr(val)
+                      for key, val in self._kwargs.items()) \
+            + ")"
 
 
 class AtspiStateTracker(EventSource):
@@ -44,10 +66,10 @@ class AtspiStateTracker(EventSource):
     _async_event_names      = ("async-focus-changed",
                                "async-text-changed",
                                "async-text-caret-moved")
-    _event_names = _async_event_names + \
-                   _focus_event_names + \
-                   _text_event_names + \
-                   _key_stroke_event_names
+    _event_names = (_async_event_names +
+                    _focus_event_names +
+                    _text_event_names +
+                    _key_stroke_event_names)
 
     _focus_listeners_registered = False
     _keystroke_listeners_registered = False
@@ -56,8 +78,10 @@ class AtspiStateTracker(EventSource):
     _keystroke_listener = None
 
     # asynchronously accessible members
-    _focused_accessible = None   # any currently focused accessible
-    _active_accessible = None    # editable focused accessible
+    _focused_accessible = None   # last focused editable accessible
+    _focused_pid = None          # pid of last focused editable accessible
+    _active_accessible = None    # currently active editable accessible
+    _active_accessible_activation_time = 0.0  # time since focus received
     _last_active_accessible = None
     _state = None                # cache of various accessible properties
 
@@ -119,7 +143,7 @@ class AtspiStateTracker(EventSource):
         self._register_atspi_keystroke_listeners(register)
 
     def _register_atspi_focus_listeners(self, register):
-        if not "Atspi" in globals():
+        if "Atspi" not in globals():
             return
 
         if self._focus_listeners_registered != register:
@@ -149,27 +173,32 @@ class AtspiStateTracker(EventSource):
             self._focus_listeners_registered = register
 
     def _register_atspi_text_listeners(self, register):
-        if not "Atspi" in globals():
+        if "Atspi" not in globals():
             return
 
         if self._text_listeners_registered != register:
             if register:
                 self.atspi_connect("_listener_text_changed",
-                                   "object:text-changed",
+                                   "object:text-changed:insert",
+                                   self._on_atspi_text_changed)
+                self.atspi_connect("_listener_text_changed",
+                                   "object:text-changed:delete",
                                    self._on_atspi_text_changed)
                 self.atspi_connect("_listener_text_caret_moved",
                                    "object:text-caret-moved",
                                    self._on_atspi_text_caret_moved)
             else:
                 self.atspi_disconnect("_listener_text_changed",
-                                      "object:text-changed")
+                                      "object:text-changed:insert")
+                self.atspi_disconnect("_listener_text_changed",
+                                      "object:text-changed:delete")
                 self.atspi_disconnect("_listener_text_caret_moved",
                                       "object:text-caret-moved")
 
         self._text_listeners_registered = register
 
     def _register_atspi_keystroke_listeners(self, register):
-        if not "Atspi" in globals():
+        if "Atspi" not in globals():
             return
 
         if self._keystroke_listeners_registered != register:
@@ -178,15 +207,16 @@ class AtspiStateTracker(EventSource):
             if register:
                 if not self._keystroke_listener:
                     self._keystroke_listener = \
-                       Atspi.DeviceListener.new(self._on_atspi_keystroke, None)
+                        Atspi.DeviceListener.new(self._on_atspi_keystroke,
+                                                 None)
 
                 for modifier_mask in modifier_masks:
-                    Atspi.register_keystroke_listener( \
-                                        self._keystroke_listener,
-                                        None,        # key set, None=all
-                                        modifier_mask,
-                                        Atspi.KeyEventType.PRESSED,
-                                        Atspi.KeyListenerSyncType.SYNCHRONOUS)
+                    Atspi.register_keystroke_listener(
+                        self._keystroke_listener,
+                        None,        # key set, None=all
+                        modifier_mask,
+                        Atspi.KeyEventType.PRESSED,
+                        Atspi.KeyListenerSyncType.SYNCHRONOUS)
             else:
                 # Apparently any single deregister call will turn off
                 # all the other registered modifier_masks too. Since
@@ -197,10 +227,10 @@ class AtspiStateTracker(EventSource):
 
                 for modifier_mask in modifier_masks:
                     Atspi.deregister_keystroke_listener(
-                                        self._keystroke_listener,
-                                        None, # key set, None=all
-                                        modifier_mask,
-                                        Atspi.KeyEventType.PRESSED)
+                        self._keystroke_listener,
+                        None,  # key set, None=all
+                        modifier_mask,
+                        Atspi.KeyEventType.PRESSED)
 
         self._keystroke_listeners_registered = register
 
@@ -246,7 +276,7 @@ class AtspiStateTracker(EventSource):
         if not self._frozen:
             EventSource.emit_async(self, event_name, *args, **kwargs)
 
-    ########## synchronous handlers ##########
+    # ######### synchronous handlers ######### #
 
     def _on_atspi_global_focus(self, event, user_data):
         self._on_atspi_focus(event, True)
@@ -254,41 +284,45 @@ class AtspiStateTracker(EventSource):
     def _on_atspi_object_focus(self, event, user_data):
         self._on_atspi_focus(event)
 
-    def _on_atspi_focus(self, event, focus_received = False):
-        focused = bool(focus_received) or bool(event.detail1) # received focus?
-        ae = AsyncEvent(accessible = event.source,
-                        focused = focused)
+    def _on_atspi_focus(self, event, focus_received=False):
+        focused = (bool(focus_received) or
+                   bool(event.detail1))  # received focus?
+        ae = AsyncEvent(accessible=event.source,
+                        focused=focused)
         self.emit_async("async-focus-changed", ae)
 
     def _on_atspi_text_changed(self, event, user_data):
-        #print("_on_atspi_text_changed", event.detail1, event.detail2, event.source, event.type, event.type.endswith("delete"))
-        ae = AsyncEvent(accessible = event.source,
-                        type = event.type,
-                        pos = event.detail1,
-                        length = event.detail2)
+        # print("_on_atspi_text_changed", event.detail1, event.detail2,
+        #       event.source, event.type, event.type.endswith("delete"))
+        ae = AsyncEvent(accessible=event.source,
+                        type=event.type,
+                        pos=event.detail1,
+                        length=event.detail2)
         self.emit_async("async-text-changed", ae)
         return False
 
     def _on_atspi_text_caret_moved(self, event, user_data):
-        #print("_on_atspi_text_caret_moved", event.detail1, event.detail2, event.source, event.type, event.source.get_name(), event.source.get_role())
-        ae = AsyncEvent(accessible = event.source,
-                        caret = event.detail1)
+        # print("_on_atspi_text_caret_moved", event.detail1, event.detail2,
+        #       event.source, event.type, event.source.get_name(),
+        #       event.source.get_role())
+        ae = AsyncEvent(accessible=event.source,
+                        caret=event.detail1)
         self.emit_async("async-text-caret-moved", ae)
         return False
 
     def _on_atspi_keystroke(self, event, user_data):
         if event.type == Atspi.EventType.KEY_PRESSED_EVENT:
-            _logger.atspi("key-stroke {} {} {} {}" \
+            _logger.atspi("key-stroke {} {} {} {}"
                           .format(event.modifiers,
                                   event.hw_code, event.id, event.is_text))
-            #keysym = event.id # What is this? Not XK_ keysyms apparently.
-            ae = AsyncEvent(hw_code   = event.hw_code,
-                            modifiers = event.modifiers)
+            # keysym = event.id # What is this? Not an XK_ keysym apparently.
+            ae = AsyncEvent(hw_code=event.hw_code,
+                            modifiers=event.modifiers)
             self.emit_async("key-pressed", ae)
 
-        return False # don't consume event
+        return False  # don't consume event
 
-    ########## asynchronous handlers ##########
+    # ######### asynchronous handlers ######### #
 
     def _on_async_focus_changed(self, event):
         accessible = event.accessible
@@ -297,71 +331,106 @@ class AtspiStateTracker(EventSource):
 
         # Don't access the accessible while frozen. This leads to deadlocks
         # while displaying Onboard's own dialogs/popup menu's.
-        if not self._frozen:
+        if accessible and not self._frozen:
             self._log_accessible(accessible, focused)
 
-            if not accessible is None:
-                try:
-                    self._state = \
-                            self._read_initial_accessible_state(accessible)
-                except Exception as ex: # Private exception gi._glib.GError when
-                                        # gedit became unresponsive.
-                    _logger.atspi("_on_focus_changed(): "
-                                 "invalid accessible, failed to read state: " \
-                                 + unicode_str(ex))
-
-                editable = self._is_accessible_editable(self._state)
-                activate = focused and editable
-
-                if focused:
-                    self._focused_accessible = accessible
-                elif not focused and self._focused_accessible == accessible:
+            # Since Trusty, focus events no longer come reliably in a
+            # predictable order. -> Store the last editable accessible
+            # so we can pick it over later focused non-editable ones.
+            # Helps to keep the keyboard open in presence of popup selections
+            # e.g. in GNOME's file dialog and in Unity Dash.
+            state_valid = False
+            ignore_accessible = False
+            if self._focused_accessible is accessible:
+                if not focused:
                     self._focused_accessible = None
-                else:
-                    activate = None
+            else:
+                if focused:
+                    self._state = \
+                        self._read_initial_accessible_state(accessible)
+                    pid = self._state.get("pid")
 
-                if not activate is None:
-                    if activate:
-                        active_accessible = self._focused_accessible
+                    if self._is_accessible_editable(self._state):
+                        self._focused_accessible = accessible
+                        self._focused_pid = pid
+                        state_valid = True
                     else:
-                        active_accessible = None
+                        # Wily: prevent random icons, buttons and toolbars
+                        # in unity dash from hiding Onboard. Somehow hovering
+                        # over those buttons silently drops the focus from the
+                        # text entry. Let's pretend the buttons don't exist
+                        # and keep the previously saved text entry active.
+                        app_name = self._state.get("app-name", "").lower()
+                        if app_name == "unity":
+                            ignore_accessible = True
+                        else:
+                            # Wily: attempt to hide when unity dash closes
+                            # (there's no focus lost event).
+                            # Also check duration since last activation to
+                            # skip out of order focus events (firefox
+                            # ATSPI_ROLE_DOCUMENT_FRAME) for a short while
+                            # after opening dash.
+                            now = time.time()
+                            if now - self._active_accessible_activation_time \
+                               > .5:
+                                if self._focused_pid != pid:
+                                    self._focused_accessible = None
+                                    _logger.atspi("Dropping accessible due to "
+                                                  "pid change: {} != {} "
+                                                  .format(self._focused_pid,
+                                                          pid))
 
-                    self._set_active_accessible(active_accessible)
+            if not ignore_accessible:
+                # Make sure we have a valid state for all cases.
+                if not state_valid:
+                    self._state = self._read_initial_accessible_state(
+                        self._focused_accessible)
+
+                # Has the previously focused accessible lost the focus?
+                active_accessible = self._focused_accessible
+                if not self._is_accessible_focused(self._state):
+                    active_accessible = None
+
+                self._set_active_accessible(active_accessible)
 
     def _set_active_accessible(self, accessible):
         self._active_accessible = accessible
 
-        if not self._active_accessible is None or \
-           not self._last_active_accessible is None:
+        if self._active_accessible is not None or \
+           self._last_active_accessible is not None:
 
-            if not accessible is None:
+            if accessible is not None:
                 try:
-                    self._state.update( \
-                                self._read_remaining_accessible_state(accessible))
-                except Exception as ex: # Private exception gi._glib.GError when
-                                        # gedit became unresponsive.
+                    self._state.update(
+                        self._read_remaining_accessible_state(accessible))
+                # Private exception gi._glib.GError when
+                # gedit became unresponsive.
+                except Exception as ex:
+
                     _logger.atspi("_set_active_accessible(): "
-                            "invalid accessible, failed to read remaining state: " \
-                            + unicode_str(ex))
+                                  "invalid accessible, failed to "
+                                  "read remaining state: " +
+                                  unicode_str(ex))
 
             # notify listeners
             self.emit("text-entry-activated",
                       self._active_accessible)
 
             self._last_active_accessible = self._active_accessible
+            self._active_accessible_activation_time = time.time()
 
     def _on_async_text_changed(self, event):
         if event.accessible is self._active_accessible:
             type = event.type
             insert = type.endswith(("insert", "insert:system"))
             delete = type.endswith(("delete", "delete:system"))
-            #print(event.accessible.get_id(), type, insert)
+            # print(event.accessible.get_id(), type, insert)
             if insert or delete:
                 event.insert = insert
                 self.emit("text-changed", event)
             else:
                 _logger.warning("_on_async_text_changed: "
-                                "unknown event type '{}'" \
+                                "unknown event type '{}'"
                                 .format(event.type))
 
     def _on_async_text_caret_moved(self, event):
@@ -389,6 +458,43 @@ class AtspiStateTracker(EventSource):
         """ Screen rect of the active accessible """
         return self._state.get("extents", Rect())
 
+    def get_frame(self):
+        if not self._active_accessible:
+            return None
+
+        frame = self._state.get("frame")
+        if not frame:
+            frame = self._get_accessible_frame(self._active_accessible)
+            self._state["frame"] = frame   # may be None
+        return frame
+
+    @staticmethod
+    def get_accessible_extents(accessible):
+        """ Screen rect of the given accessible, no caching """
+        try:
+            rect = AtspiStateTracker._get_accessible_extents(accessible)
+        # private exception gi._glib.GError when
+        # right clicking onboards unity2d launcher (Precise)
+        except Exception as ex:
+            _logger.atspi("Invalid accessible,"
+                          " failed to get extents: " + unicode_str(ex))
+            rect = Rect()
+        return rect
+
+    @staticmethod
+    def get_accessible_character_extents(accessible, offset):
+        """ Screen rect of the character at offset of the accessible """
+        try:
+            rect = AtspiStateTracker._get_accessible_character_extents(
+                accessible, offset)
+        except Exception as ex:  # private exception gi._glib.GError when
+                # right clicking onboards unity2d launcher (Precise)
+            _logger.atspi("Invalid accessible,"
+                          " failed to get character extents: " +
+                          unicode_str(ex))
+            rect = Rect()
+        return rect
+
     @staticmethod
     def _get_accessible_extents(accessible):
         """
@@ -411,36 +517,76 @@ class AtspiStateTracker(EventSource):
                     ext.width * scale, ext.height * scale)
 
     @staticmethod
-    def get_accessible_extents(accessible):
-        """ Screen rect of the given accessible, no caching """
-        try:
-            rect = AtspiStateTracker._get_accessible_extents(accessible)
-        except Exception as ex: # private exception gi._glib.GError when
-                # right clicking onboards unity2d launcher (Precise)
-            _logger.atspi("Invalid accessible,"
-                         " failed to get extents: " + unicode_str(ex))
-            rect = Rect()
-        return rect
+    def _get_accessible_character_extents(accessible, offset):
+        """
+        Screen rect of the character at offset of the accessible, no caching,
+        no exception handling.
+        """
+        scale = config.window_scaling_factor
+        if scale != 1.0:
+            attributes = accessible.get_attributes()
+            # Only Gtk-3 widgets return scaled coordinates, all others,
+            # including Gtk-2 apps like firefox, clawsmail and Qt-apps,
+            # apparently don't.
+            if AtspiStateTracker.is_toolkit_gtk3(attributes):
+                scale = 1.0
+            else:
+                scale = 1.0 / config.window_scaling_factor
+
+        ext = accessible.get_character_extents(offset, Atspi.CoordType.SCREEN)
+        # x, y = ext.x + ext.width / 2, ext.y + ext.height / 2
+        # offset_control = accessible.get_offset_at_point(x, y,
+        #                                                Atspi.CoordType.SCREEN)
+        # print(offset, offset_control)
+        return Rect(ext.x * scale, ext.y * scale,
+                    ext.width * scale, ext.height * scale)
 
     @staticmethod
     def get_accessible_text(accessible, begin, end):
         """ Text of the given accessible, no caching """
         try:
             text = Atspi.Text.get_text(accessible, begin, end)
-        except Exception as ex: # private exception
-                                # gi._glib.GError: timeout from dbind
-                                # with web search in firefox.
+        # private exception gi._glib.GError: timeout from dbind
+        # with web search in firefox.
+        except Exception as ex:
             _logger.atspi("Invalid accessible,"
-                         " failed to get text: " + unicode_str(ex))
+                          " failed to get text: " + unicode_str(ex))
             return None
 
         return text
 
+    @staticmethod
+    def _get_accessible_frame(accessible):
+        """ Accessible of the top level window to which accessible belongs. """
+        frame = None
+        _logger.atspi("_get_accessible_frame(): searching for top level:")
+        try:
+            parent = accessible
+            while True:
+                parent = parent.get_parent()
+                if not parent:
+                    break
+                role = parent.get_role()
+                _logger.atspi("parent: {}".format(role))
+                if role == Atspi.Role.FRAME or \
+                   role == Atspi.Role.DIALOG or \
+                   role == Atspi.Role.WINDOW or \
+                   role == Atspi.Role.NOTIFICATION:
+                    frame = parent
+                    break
+        # private exception gi._glib.GError when
+        # right clicking onboards unity2d launcher (Precise)
+        except Exception as ex:
+            _logger.atspi("Invalid accessible,"
+                          " failed to get top level accessible: " +
+                          unicode_str(ex))
+        return frame
+
     def _is_accessible_editable(self, acc_state):
         """ Is this an accessible onboard should be shown for? """
         role      = acc_state.get("role")
         state_set = acc_state.get("state-set")
-        if not state_set is None:
+        if state_set is not None:
 
             if role in [Atspi.Role.TEXT,
                         Atspi.Role.TERMINAL,
@@ -457,26 +603,60 @@ class AtspiStateTracker(EventSource):
                         Atspi.Role.PARAGRAPH,      # LibreOffice Writer
                         Atspi.Role.HEADER,
                         Atspi.Role.FOOTER,
-                       ]:
+                        ]:
                 if role in [Atspi.Role.TERMINAL] or \
-                   (not state_set is None and \
+                   (state_set is not None and
                     state_set.contains(Atspi.StateType.EDITABLE)):
                     return True
         return False
 
+    def _is_accessible_focused(self, state):
+        state_set = state.get("state-set")
+        if state_set:
+            return state_set.contains(Atspi.StateType.FOCUSED)
+        return False
+
     def _read_initial_accessible_state(self, accessible):
         """
-        Read just enough to find out if we are intereseted in this accessible.
+        Read just enough to find out if we are interested in this accessible.
         """
         state = {}
-        state["role"] = accessible.get_role()
-        state["state-set"] = accessible.get_state_set()
+        if accessible is not None:
+            try:
+                state["role"] = accessible.get_role()
+                state["state-set"] = accessible.get_state_set()
+                state["id"] = accessible.get_id()
+            # Private exception gi._glib.GError when
+            # gedit became unresponsive.
+            except Exception as ex:
+                _logger.info("_read_initial_accessible_state(): "
+                             "invalid accessible, failed to read state: " +
+                             unicode_str(ex))
+            try:
+                state["pid"] = accessible.get_process_id()
+            # Private exception gi._glib.GError when
+            # gedit became unresponsive.
+            except Exception as ex:
+
+                _logger.info("_read_initial_accessible_state(): "
+                             "failed to get pid: " +
+                             unicode_str(ex))
+            try:
+                app = accessible.get_application()
+                state["app-name"] = app.get_name()
+            # Private exception gi._glib.GError when
+            # gedit became unresponsive.
+            except Exception as ex:
+
+                _logger.info("_read_initial_accessible_state(): "
+                             "failed to get app-name: " +
+                             unicode_str(ex))
         return state
 
     def _read_remaining_accessible_state(self, accessible):
         """
         Read more attributes and find out as much as we
-        can about the accessibles purpose.
+        can about the accessible's purpose.
         """
         state = {}
 
@@ -484,12 +664,12 @@ class AtspiStateTracker(EventSource):
         state["interfaces"] = accessible.get_interfaces()
         state["extents"] = self._get_accessible_extents(accessible)
 
-        # These are currently used only in debug output
+        # These are currently used for debug output only
         if _logger.isEnabledFor(_logger.LEVEL_ATSPI):
             state["id"] = accessible.get_id()
             state["name"] = accessible.get_name()
             pid = accessible.get_process_id()
-            state["process-id"] = pid
+            state["pid"] = pid
             if pid != -1:
                 state["process-name"] = Process.get_process_name(pid)
 
@@ -498,13 +678,28 @@ class AtspiStateTracker(EventSource):
                 state["app-name"] = app.get_name()
                 state["app-description"] = app.get_description()
 
+            state["toolkit-name"] = accessible.get_toolkit_name()
+            state["toolkit-version"] = accessible.get_toolkit_version()
+            # state["summary"] = accessible.get_summary()
+            state["editable_text_iface"] = accessible.get_editable_text_iface()
+            # state["document_attributes"] = \
+            #    accessible.get_document_attributes()
+            state["description"] = accessible.get_description()
+            # state["default_attributes"] = \
+            #    accessible.get_default_attributes() # not impl. by unity dash
+
+            frame = self._get_accessible_frame(accessible)
+            state["frame"] = frame
+            state["frame_extents"] = self._get_accessible_extents(frame) \
+                if frame else None
+
         return state
 
     @staticmethod
     def is_toolkit_gtk3(attributes):
         """ Are the accessible attributes from a gtk3 widget? """
         return attributes and \
-               "toolkit" in attributes and attributes["toolkit"] == "gtk"
+            "toolkit" in attributes and attributes["toolkit"] == "gtk"
 
     def _log_accessible(self, accessible, focused):
         if _logger.isEnabledFor(_logger.LEVEL_ATSPI):
@@ -526,20 +721,22 @@ class AtspiStateTracker(EventSource):
                     state_set = accessible.get_state_set()
                     states = state_set.states
                     editable = state_set.contains(Atspi.StateType.EDITABLE) \
-                               if state_set else None
+                        if state_set else None
                     extents = self._get_accessible_extents(accessible)
-                except: # private exception gi._glib.GError when gedit became unresponsive
+                # private exception gi._glib.GError when gedit became
+                # unresponsive
+                except:
                     pass
 
                 msg += "name={name}, role={role}({role_name}), " \
                        "editable={editable}, states={states}, " \
                        "extents={extents}]" \
-                        .format(name=name,
-                                role = role.value_name if role else role,
-                                role_name = role_name,
-                                editable = editable,
-                                states = states,
-                                extents = extents \
+                       .format(name=name,
+                               role=role.value_name if role else role,
+                               role_name=role_name,
+                               editable=editable,
+                               states=states,
+                               extents=extents
                                )
             _logger.atspi(msg)
 
@@ -587,14 +784,15 @@ class AtspiStateType:
               'VERTICAL',
               'VISIBLE',
               'VISITED',
-             ]
+              ]
 
     @staticmethod
     def to_strings(state_set):
         result = []
-        for s in AtspiStateType.states:
-            if state_set.contains(getattr(Atspi.StateType, s)):
-                result.append(s)
+        if state_set is not None:
+            for s in AtspiStateType.states:
+                if state_set.contains(getattr(Atspi.StateType, s)):
+                    result.append(s)
         return result
 
 
