@@ -1,9 +1,9 @@
 # -*- coding: utf-8 -*-
 
 # Copyright © 2008-2010 Chris Jones <tortoise@tortuga>
-# Copyright © 2009-2010 Francesco Fumanti <francesco.fumanti@gmx.net>
+# Copyright © 2008-2011 Francesco Fumanti <francesco.fumanti@gmx.net>
 # Copyright © 2012 Gerd Kohlberger <lowfi@chello.at>
-# Copyright © 2009, 2011-2015 marmuta <marmvta@gmail.com>
+# Copyright © 2009, 2011-2016 marmuta <marmvta@gmail.com>
 #
 # This file is part of Onboard.
 #
@@ -40,7 +40,8 @@ from Onboard.WindowUtils    import show_confirmation_dialog
 from Onboard.utils          import Version, \
                                    unicode_str, XDGDirs, chmodtree, \
                                    Process, hexcolor_to_rgba, TermColors
-from Onboard.definitions    import StatusIconProviderEnum, \
+from Onboard.definitions    import DesktopEnvironmentEnum, \
+                                   StatusIconProviderEnum, \
                                    InputEventSourceEnum, \
                                    TouchInputEnum, \
                                    LearningBehavior, \
@@ -106,7 +107,7 @@ USER_DIR                   = "onboard"
 
 SYSTEM_DEFAULTS_FILENAME   = "onboard-defaults.conf"
 
-DEFAULT_WINDOW_HANDLES     = list(Handle.ALL)
+DEFAULT_WINDOW_HANDLES     = list(Handle.RESIZE_MOVE)
 
 DEFAULT_FREQUENCY_TIME_RATIO = 75  # 0=100% frequency, 100=100% time (last use)
 
@@ -201,6 +202,8 @@ class Config(ConfigObject):
     _xembed_background_rgba = None
     _xembed_background_image_enabled = None
 
+    _desktop_environment = None
+
     def __new__(cls, *args, **kwargs):
         """
         Singleton magic.
@@ -250,6 +253,10 @@ class Config(ConfigObject):
                        "environments. DESKTOPS is a comma-separated list of "
                        "XDG desktop names, e.g. GNOME for GNOME Shell."
                        ))
+        group.add_option("-D", "--startup-delay",
+                type="float", dest="startup_delay",
+                help=_("Delay showing the initial keyboard window "
+                       "by STARTUP_DELAY seconds (default 0.0)."))
         group.add_option("-a", "--keep-aspect", action="store_true",
                 dest="keep_aspect_ratio",
                 help=_("Keep aspect ratio when resizing the window"))
@@ -286,6 +293,7 @@ class Config(ConfigObject):
         self.log_learn = options.log_learn
         self.quirks_name = options.quirks_name
         self.quirks = None  # WMQuirks instance, provided by KbdWindow for now
+        self.startup_delay = options.startup_delay
 
         # optionally log to file; everything, including stack traces
         if 0:
@@ -341,6 +349,10 @@ class Config(ConfigObject):
             _logger.error(unicode_str(e))
             sys.exit()
 
+        # log desktop environment
+        _logger.debug("Desktop environment: {}"
+                      .format(self.get_desktop_environment().name))
+
         # log how we were launched
         if _logger.isEnabledFor(logging.DEBUG):
             _logger.debug("command line: " + str(sys.argv))
@@ -605,8 +617,9 @@ class Config(ConfigObject):
         self.add_key("snippets", {}, "as")
         self.add_key("show-status-icon", True)
         self.add_key("status-icon-provider", StatusIconProviderEnum.AppIndicator,
-                                             enum={"GtkStatusIcon" : 0,
-                                                   "AppIndicator" : 1,
+                                             enum={"auto" : 0,
+                                                   "GtkStatusIcon" : 1,
+                                                   "AppIndicator" : 2,
                                                   })
         self.add_key("start-minimized", False)
         self.add_key("xembed-onboard", False, prop="onboard_xembed_enabled")
@@ -1164,11 +1177,35 @@ class Config(ConfigObject):
         return self.window.enable_inactive_transparency and \
                not self.scanner.enabled
 
-    def is_keep_aspect_ratio_enabled(self):
-        return self.window.keep_aspect_ratio or \
-               self.options.keep_aspect_ratio or \
-               (self.xid_mode and self.launched_by != self.LAUNCHER_NONE and \
-                not self.system_defaults.get("xembed_aspect_change_range") is None)
+    def is_keep_window_aspect_ratio_enabled(self, orientation_co):
+        """
+        Keep aspect ratio of the whole keyboard window?
+        Not recommended, no effect in MATE and elsewhere the
+        keyboard tends to shrink with each interaction.
+        """
+        return ((self.window.keep_aspect_ratio or
+                 self.options.keep_aspect_ratio) and
+                not self.xid_mode and
+                not self.is_docking_enabled())
+
+    def is_keep_frame_aspect_ratio_enabled(self, orientation_co):
+        """
+        Keep aspect ratio of only the frame (keyboard area)
+        inside the keyboard window?
+        """
+        return \
+            self.is_keep_xembed_frame_aspect_ratio_enabled() or \
+            self.is_keep_docking_frame_aspect_ratio_enabled(orientation_co)
+
+    def is_keep_xembed_frame_aspect_ratio_enabled(self):
+        return (self.xid_mode and self.launched_by != self.LAUNCHER_NONE and
+                self.system_defaults.get("xembed_aspect_change_range")
+                is not None)
+
+    def is_keep_docking_frame_aspect_ratio_enabled(self, orientation_co):
+        return (not self.xid_mode and
+                self.is_docking_enabled() and
+                self.is_dock_expanded(orientation_co))
 
     def is_mousetweaks_active(self):
         return self.mousetweaks and self.mousetweaks.is_active()
@@ -1191,8 +1228,8 @@ class Config(ConfigObject):
 
     def set_num_window_handles(self, num):
         if num == NumResizeHandles.ALL:
-            window_handles = list(Handle.ALL)
-            icp_handles    = list(Handle.ALL)
+            window_handles = list(Handle.RESIZE_MOVE)
+            icp_handles    = list(Handle.RESIZE_MOVE)
         elif num == NumResizeHandles.NORESIZE:
             window_handles = [Handle.MOVE]
             icp_handles    = [Handle.MOVE]
@@ -1418,11 +1455,85 @@ class Config(ConfigObject):
             lang_id = "en_US"
         return lang_id
 
+    def get_desktop_environment(self):
+        """
+        Return current desktop environment.
+        """
+        if not self._desktop_environment:
+            self._desktop_environment = self._detect_desktop_environment()
+        return self._desktop_environment
+
+    @staticmethod
+    def _detect_desktop_environment():
+        """
+        Detect current desktop environment. Extend as needed.
+        """
+        xdg_desktop = os.environ.get("XDG_CURRENT_DESKTOP", "")
+        desktop = os.environ.get("DESKTOP_SESSION", "")
+
+        def istrcmp(s1, s2):
+            if sys.version_info < (3,3):
+                return s1.lower() == s2.lower()
+            return s1.casefold() == s2.casefold()
+
+
+        def isdesktop(id):
+            return istrcmp(xdg_desktop, id) or istrcmp(desktop, id)
+
+        if isdesktop("X-Cinnamon") or isdesktop("cinnamon"):
+            return DesktopEnvironmentEnum.Cinnamon
+        if isdesktop("GNOME"):
+            return DesktopEnvironmentEnum.GNOME_Shell
+        if isdesktop("GNOME-Classic:GNOME"):
+            return DesktopEnvironmentEnum.GNOME_Classic
+        if isdesktop("KDE"):
+            return DesktopEnvironmentEnum.KDE
+        if isdesktop("lxde"):
+            return DesktopEnvironmentEnum.LXDE
+        if isdesktop("LXQt"):
+            return DesktopEnvironmentEnum.LXQT
+        if isdesktop("mate"):
+            return DesktopEnvironmentEnum.MATE
+        if isdesktop("Unity"):
+            return DesktopEnvironmentEnum.Unity
+        if isdesktop("xfce"):
+            return DesktopEnvironmentEnum.XFCE
+
+        return DesktopEnvironmentEnum.Unknown
+
+    def prefer_gtkstatusicon(self):
+        """
+        Auto-detect if we should fall back to GtkStatusIcon.
+        """
+        de = self.get_desktop_environment()
+        if de in (
+            # AppIndicator is supported in XUbuntu 16.04, but w/o left click
+            # activation. GtkStatusIcon works well and allows left click.
+            DesktopEnvironmentEnum.Cinnamon,
+
+            # AppIndicator does nothing in 16.04.
+            DesktopEnvironmentEnum.MATE,
+
+            # AppIndicator does nothing in 16.04.
+            DesktopEnvironmentEnum.XFCE,
+
+            # AppIndicator is supported in Lubuntu 16.04, but w/o left
+            # click activation. GtkStatusIcon works well, but is placed
+            # into a legacy system tray applet with tiny 16x16 icon.
+            # DesktopEnvironmentEnum.LXDE,
+
+            # AppIndicator is supported in LXQt 16.04, including left
+            # click activation. GtkStatusIcon works well too.
+            # DesktopEnvironmentEnum.LXQT,
+        ):
+            return True
+        return False
+
 
 class ConfigKeyboard(ConfigObject):
     """Window configuration """
     DEFAULT_KEY_ACTION = 1 # Release-only, supports long press
-    DEFAULT_KEY_SYNTH  = 0 # XTest
+    DEFAULT_KEY_SYNTH  = 0 # Auto
     DEFAULT_TOUCH_INPUT = TouchInputEnum.MULTI
     DEFAULT_INPUT_EVENT_SOURCE = InputEventSourceEnum.XINPUT
 
@@ -1440,8 +1551,10 @@ class ConfigKeyboard(ConfigObject):
                                                  "delayed-stroke" : 1,
                                                 })
         self.add_key("key-synth", self.DEFAULT_KEY_SYNTH,
-                                           enum={"XTest" : 0,
-                                                 "AT-SPI" : 1,
+                                           enum={"auto" : 0,
+                                                 "XTest" : 1,
+                                                 "uinput" : 2,
+                                                 "AT-SPI" : 3,
                                                 })
         self.add_key("touch-input", self.DEFAULT_TOUCH_INPUT,
                                            enum={"none" : 0,
@@ -1520,6 +1633,7 @@ class ConfigWindow(ConfigObject):
                                            "monitor8" : DockingMonitor.MONITOR8,
                                           })
         self.add_key("docking-shrink-workarea", True)
+        self.add_key("docking-aspect-change-range", [0, 1.6], "ad")
 
         self.landscape = ConfigWindow.Landscape(self)
         self.portrait = ConfigWindow.Portrait(self)
@@ -1591,7 +1705,7 @@ class ConfigWindow(ConfigObject):
             self.add_key("height", DEFAULT_HEIGHT)
             self.add_key("dock-width", DEFAULT_WIDTH)
             self.add_key("dock-height", DEFAULT_HEIGHT)
-            self.add_key("dock-expand", False)
+            self.add_key("dock-expand", True)
 
     class Portrait(ConfigObject):
         def _init_keys(self):
@@ -1776,15 +1890,34 @@ class ConfigTheme(ConfigObject):
 
         return value
 
+    _font_attributes = ("bold", "italic", "condensed")
+    _cached_key_label_font = None
+
+    def _post_notify_key_label_font(self):
+        self._cached_key_label_font = None
+
     def get_key_label_font(self):
-        gskey = self.key_label_font_key
+        if self._cached_key_label_font is None:
+            gskey = self.key_label_font_key
 
-        value = gskey.value
-        if not value:
-            # get default value from onboard base config instead
-            value = self.parent.key_label_font
+            value = gskey.value
+            if value:
+                items = value.split()
+                if items:
+                    # If no font family was provided merge in system font family
+                    if items[0] in self._font_attributes:
+                        system_items = self.parent.key_label_font.split()
+                        if system_items and \
+                        system_items[0] not in self._font_attributes:
+                            items.insert(0, system_items[0])
+                            value = " ".join(items)
+            else:
+                # get default value from onboard base config instead
+                value = self.parent.key_label_font
 
-        return value
+            self._cached_key_label_font = value
+
+        return self._cached_key_label_font
 
 
 class ConfigLockdown(ConfigObject):
