@@ -3,7 +3,7 @@
 # Copyright © 2008 Chris Jones <tortoise@tortuga>
 # Copyright © 2009-2011 Francesco Fumanti <francesco.fumanti@gmx.net>
 # Copyright © 2012 Gerd Kohlberger <lowfi@chello.at>
-# Copyright © 2010-2015 marmuta <marmvta@gmail.com>
+# Copyright © 2010-2016 marmuta <marmvta@gmail.com>
 #
 # This file is part of Onboard.
 #
@@ -32,7 +32,7 @@ from Onboard.definitions import DockingMonitor
 from Onboard.utils       import Rect, Version
 from Onboard.Timer       import CallOnce, Timer
 from Onboard.WindowUtils import Orientation, WindowRectPersist, \
-                                set_unity_property, \
+                                WindowRectTracker, set_unity_property, \
                                 gtk_has_resize_grip_support
 import Onboard.osk as osk
 
@@ -111,6 +111,7 @@ class KbdWindowBase:
         self.connect('composited-changed',      self._cb_composited_changed)
         self.connect("realize",                 self._cb_realize_event)
         self.connect("unrealize",               self._cb_unrealize_event)
+        self.connect("configure-event",         self._on_configure_event)
 
         self.detect_window_manager()
         self.check_alpha_support()
@@ -327,6 +328,49 @@ class KbdWindowBase:
     def update_taskbar_hint(self):
         config.quirks.update_taskbar_hint(self)
 
+    def _on_configure_event(self, widget, event):
+        self.update_window_rect()
+        self.update_window_scaling_factor()
+
+        if _logger.isEnabledFor(logging.DEBUG):
+            _logger.debug(
+                "configure-event: event {}, allocated {}, "
+                "window {}, scaling factor {}"
+                .format(Rect(event.x, event.y, event.width, event.height),
+                        (self.get_allocated_width(),
+                        self.get_allocated_height()),
+                        self.get_rect(),
+                        config.window_scaling_factor))
+
+    def update_window_scaling_factor(self, startup = False):
+        """ check for changes to the window-scaling-factor """
+        scale = self.get_scale_factor()
+        _logger.debug("update_window_scaling_factor: scale {}"
+                      .format(scale))
+
+        if not scale is None and \
+           config.window_scaling_factor != scale:
+            config.window_scaling_factor = scale
+            _logger.info("new window-scaling-factor '{}'".format(scale))
+
+            keyboard = self.keyboard_widget.keyboard
+
+            # In override redirect mode the window gets stuck with
+            # the old scale -> recreate it
+            if config.is_override_redirect() and \
+               not startup and \
+               not config.xid_mode:
+                self._recreate_window()
+
+                # Release pressed keys in case the switch was initiated
+                # by Onboard, e.g. with the Return key in the terminal.
+                keyboard.release_pressed_keys()
+
+            # invalidate all images and key surfaces
+            self.keyboard_widget.invalidate_images()
+            keyboard.invalidate_ui()
+            keyboard.commit_ui_updates()
+
     def is_visible(self):
         if not self.get_mapped():
             return False
@@ -415,7 +459,7 @@ class KbdWindowBase:
         Fails to be called when iconifying in unity (Precise).
         Still keep it around for sticky changes.
         """
-        _logger.debug("window_state_event: {}, {}" \
+        _logger.debug("window-state-event: {}, {}"
                       .format(event.changed_mask, event.new_window_state))
 
         if event.changed_mask & Gdk.WindowState.MAXIMIZED:
@@ -525,6 +569,10 @@ class KbdWindowBase:
            not config.is_docking_enabled()
 
 
+    def get_orientation_config_object(self):
+        return None
+
+
 class KbdWindow(KbdWindowBase, WindowRectPersist, Gtk.Window):
 
     # Minimum window size (for resizing in system mode, see handle_motion())
@@ -536,6 +584,7 @@ class KbdWindow(KbdWindowBase, WindowRectPersist, Gtk.Window):
         self._last_ignore_configure_time = None
         self._last_configures = []
         self._was_visible = False
+        self._user_positioning_begin_rect = Rect()
 
         Gtk.Window.__init__(self,
                             urgency_hint = False,
@@ -557,7 +606,6 @@ class KbdWindow(KbdWindowBase, WindowRectPersist, Gtk.Window):
         self.connect("map",                     self._on_map_event)
         self.connect("unmap",                   self._on_unmap_event)
         self.connect("delete-event", self._on_delete_event)
-        self.connect("configure-event", self._on_configure_event)
         # Connect_after seems broken in Quantal, the callback is never called.
         #self.connect_after("configure-event", self._on_configure_event_after)
 
@@ -674,6 +722,8 @@ class KbdWindow(KbdWindowBase, WindowRectPersist, Gtk.Window):
         keyboard.stop_raise_attempts()
         keyboard.freeze_auto_show()
 
+        self._user_positioning_begin_rect = self.get_rect()
+
     def on_user_positioning_done(self):
         self.update_window_rect()
 
@@ -683,9 +733,19 @@ class KbdWindow(KbdWindowBase, WindowRectPersist, Gtk.Window):
                                     self.get_size())
             self.update_docking()
         else:
-            self.update_home_rect()
+            # Attempt to filter out accidental resize frame clicks to prevent
+            # losing the home position without clear user intention.
+            proximity = 30
+            rh = self.get_home_rect()
+            began_near_home = \
+                self._user_positioning_begin_rect.inflate(proximity) \
+                .intersection(rh) == rh
+            moved = self.keyboard_widget.was_moving()
+            update_home_position = began_near_home or moved
 
-        # Thaw auto show after a short delay to stop the window
+            self.update_home_rect(update_home_position)
+
+        # Thaw auto show only after a short delay to stop the window
         # from hiding due to spurios focus events after a system resize.
         keyboard = self.keyboard_widget.keyboard
         keyboard.thaw_auto_show(1.0)
@@ -694,35 +754,8 @@ class KbdWindow(KbdWindowBase, WindowRectPersist, Gtk.Window):
         if self.keyboard_widget.was_moving():
             config.window.docking_enabled = False
 
-    def update_window_scaling_factor(self, startup = False):
-        """ check for changes to the window-scaling-factor """
-        scale = self.get_scale_factor()
-        if not scale is None and \
-           config.window_scaling_factor != scale:
-            config.window_scaling_factor = scale
-            _logger.info("new window-scaling-factor '{}'".format(scale))
-
-            keyboard = self.keyboard_widget.keyboard
-
-            # In override redirect mode the window gets stuck with
-            # the old scale -> recreate it
-            if config.is_override_redirect() and \
-               not startup and \
-               not config.xid_mode:
-                self._recreate_window()
-
-                # Release pressed keys in case the switch was initiated
-                # by Onboard, e.g. with the Return key in the terminal.
-                keyboard.release_pressed_keys()
-
-            # invalidate all images and key surfaces
-            self.keyboard_widget.invalidate_images()
-            keyboard.invalidate_ui()
-            keyboard.commit_ui_updates()
-
     def _on_configure_event(self, widget, event):
-        self.update_window_rect()
-        self.update_window_scaling_factor()
+        KbdWindowBase._on_configure_event(self, widget, event)
 
         if not config.is_docking_enabled():
             # Connect_after seems broken in Quantal, but we still need to
@@ -887,12 +920,16 @@ class KbdWindow(KbdWindowBase, WindowRectPersist, Gtk.Window):
         if _x != x or _y != y:
             self.update_home_rect()
 
-    def update_home_rect(self):
+    def update_home_rect(self, update_home_position=True):
         if config.is_docking_enabled():
             return
 
         # update home rect
         rect = self._window_rect.copy()
+        if not update_home_position:
+            rh = self.home_rect
+            rect.x = rh.x
+            rect.y = rh.y
 
         # Make sure the move button stays visible
         if self.can_move_into_view():
@@ -1261,7 +1298,11 @@ class KbdWindow(KbdWindowBase, WindowRectPersist, Gtk.Window):
             return
 
         win = self.get_window()
-        xid = win.get_xid()  # requires GdkX11 import
+        try:
+            xid = win.get_xid()  # requires GdkX11 import
+        except AttributeError:
+            # docking unavailable on Wayland
+            return
 
         if not enable:
             self._apply_struts(xid, None)
@@ -1438,12 +1479,15 @@ class KbdWindow(KbdWindowBase, WindowRectPersist, Gtk.Window):
             win.raise_()
 
 
-class KbdPlugWindow(KbdWindowBase, Gtk.Plug):
+class KbdPlugWindow(KbdWindowBase, WindowRectTracker, Gtk.Plug):
     def __init__(self, keyboard_widget, icp = None):
         Gtk.Plug.__init__(self)
-
+        WindowRectTracker.__init__(self)
         KbdWindowBase.__init__(self, keyboard_widget, icp)
 
+    def _on_configure_event(self, widget, event):
+        KbdWindowBase._on_configure_event(self, widget, event)
+
     def toggle_visible(self):
         pass
 
