@@ -1,18 +1,40 @@
 # -*- coding: utf-8 -*-
+
+# Copyright © 2013-2016 marmuta <marmvta@gmail.com>
+#
+# This file is part of Onboard.
+#
+# Onboard is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 3 of the License, or
+# (at your option) any later version.
+#
+# Onboard is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program. If not, see <http://www.gnu.org/licenses/>.
+
 """ GTK keyboard widget """
 
 from __future__ import division, print_function, unicode_literals
 
 import cairo
 
+from Onboard.Version import require_gi_versions
+require_gi_versions()
 from gi.repository          import Gdk, Gtk, Pango, PangoCairo
 
-from Onboard.utils          import Rect, Timer, roundrect_arc
+from Onboard.utils          import Rect, roundrect_arc, Version
+from Onboard.Timer          import Timer
 from Onboard.WindowUtils    import limit_window_position, \
                                    get_monitor_rects, \
                                    canvas_to_root_window_rect, \
                                    get_monitor_dimensions, \
-                                   WindowRectTracker
+                                   WindowRectTracker, \
+                                   gtk_has_resize_grip_support
 from Onboard.TouchInput     import TouchInput
 from Onboard                import KeyCommon
 from Onboard.Layout         import LayoutRoot, LayoutPanel
@@ -82,6 +104,12 @@ class TouchFeedback:
                 popup.set_key(None)
                 del self._visible_key_feedback_popups[_key]
 
+    def raise_all(self):
+        for key, popup in self._visible_key_feedback_popups.items():
+            win = popup.get_window()
+            if win:
+                win.raise_()
+
     def _get_free_key_feedback_popup(self):
         """ Get a currently unused one from the pool of popups. """
         for popup in self._key_feedback_popup_pool:
@@ -104,7 +132,10 @@ class TouchFeedback:
 
                     # scale for hires displays
                     if gdk_win:
-                        default_size_mm *= gdk_win.get_scale_factor()
+                        try:
+                            default_size_mm *= gdk_win.get_scale_factor() # from Gdk 3.10
+                        except AttributeError:
+                            pass
 
                     w = sz[0] * default_size_mm / sz_mm[0]
                 else:
@@ -119,28 +150,66 @@ class KeyboardPopup(WindowRectTracker, Gtk.Window):
     """ Abstract base class for popups. """
 
     def __init__(self):
+        self._opacity = 1.0
+
         WindowRectTracker.__init__(self)
-        Gtk.Window.__init__(self,
-                            skip_taskbar_hint=True,
-                            skip_pager_hint=True,
-                            has_resize_grip=False,
-                            urgency_hint=False,
-                            decorated=False,
-                            accept_focus=False,
-                            opacity=1.0)
+
+        args = {
+            "skip_taskbar_hint" : True,
+            "skip_pager_hint" : True,
+            "urgency_hint" : False,
+            "decorated" : False,
+            "accept_focus" : False,
+            "opacity" : 1.0,
+            "app_paintable" : True,
+        }
+        if gtk_has_resize_grip_support():
+            args["has_resize_grip"] = False
+
+        Gtk.Window.__init__(self, **args)
 
         self.set_keep_above(True)
 
+        # In Precise, directly drawing on the top level window has no effect.
+        # The Cairo target surface is correctly rendered, but somehow it
+        # doesn't become visible. Compositing or not doesn't matter.
+        # It's most likely an old issue with Gtk/Gdk. Later releases like
+        # Trusty, Vivid are unaffected.
+        # -> Create a widget we can successfully draw on anywhere.
+        self.drawing_area = Gtk.DrawingArea()
+        self.add(self.drawing_area)
+        self.drawing_area.connect("draw", self.on_draw)
+
         # use transparency if available
         screen = Gdk.Screen.get_default()
         visual = screen.get_rgba_visual()
         self.supports_alpha = False
         if visual:
             self.set_visual(visual)
-            self.override_background_color(Gtk.StateFlags.NORMAL,
-                                           Gdk.RGBA(0, 0, 0, 0))
+            self.drawing_area.set_visual(visual)
+
+            # Somehow Gtk 3.4 still needs these now deprecated calls
+            # for LayoutPopups, even though IconPalette and LabelPopups
+            # don't. Otherwise there will be a white background.
+            gtk_version = Version(Gtk.MAJOR_VERSION, Gtk.MINOR_VERSION)
+            if gtk_version < Version(3, 18):  # Xenial doesn't need them
+                self.override_background_color(
+                    Gtk.StateFlags.NORMAL, Gdk.RGBA(0, 0, 0, 0))
+                self.drawing_area.override_background_color(
+                    Gtk.StateFlags.NORMAL, Gdk.RGBA(0, 0, 0, 0))
             self.supports_alpha = True
 
+    def set_opacity(self, opacity):
+        """ Override deprecated Gtk function of the same name """
+        self._opacity = opacity
+
+    def get_opacity(self):
+        """ Override deprecated Gtk function of the same name """
+        return self._opacity
+
+    def on_draw(self, widget, context):
+        raise NotImplementedError()
+
     def position_at(self, x, y, x_align, y_align):
         """
         Align the window with the given point.
@@ -153,13 +222,6 @@ class KeyboardPopup(WindowRectTracker, Gtk.Window):
 
         self.move(x, y)
 
-    def limit_to_workarea(self, rect, x_mon, y_mon):
-        screen = self.get_screen()
-        mon = screen.get_monitor_at_point(x_mon, y_mon)
-        area = screen.get_monitor_workarea(mon)
-        area = Rect(area.x, area.y, area.width, area.height)
-        return rect.intersection(area)
-
     def limit_to_workarea(self, rect):
         visible_rect = Rect(0, 0, rect.w, rect.h)
 
@@ -182,7 +244,6 @@ class LabelPopup(KeyboardPopup):
         KeyboardPopup.__init__(self)
         self._key = None
         self.connect("realize", self._on_realize_event)
-        self.connect("draw", self._on_draw)
 
     def _on_realize_event(self, user_data):
         self.set_override_redirect(True)
@@ -191,8 +252,7 @@ class LabelPopup(KeyboardPopup):
         win = self.get_window()
         self._osk_util.set_input_rect(win, 0, 0, 1, 1)
 
-    def _on_draw(self, widget, context):
-
+    def on_draw(self, widget, context):
         if not LabelPopup._pango_layout:
             LabelPopup._pango_layout = Pango.Layout(context=Gdk.pango_context_get())
 
@@ -213,6 +273,8 @@ class LabelPopup(KeyboardPopup):
         context.paint()
         context.restore()
 
+        context.push_group()
+
         context.set_source_rgba(*fill)
         roundrect_arc(context, content_rect, config.CORNER_RADIUS)
         context.fill()
@@ -236,6 +298,9 @@ class LabelPopup(KeyboardPopup):
                     label = "␣"
                 self._draw_text(context, label, label_rect, label_color)
 
+        context.pop_group_to_source()
+        context.paint_with_alpha(self._opacity)
+
     def _draw_text(self, context, text, rect, rgba):
         layout = self._pango_layout
         layout.set_text(text, -1)
@@ -305,8 +370,7 @@ class LayoutPopup(KeyboardPopup, LayoutView, TouchInput):
         LayoutView.__init__(self, keyboard)
         TouchInput.__init__(self)
 
-        self.connect("draw",                 self._on_draw)
-        self.connect("destroy",              self._on_destroy_event)
+        self.connect("destroy", self._on_destroy_event)
 
         self._close_timer = Timer()
         self.start_close_timer()
@@ -389,8 +453,13 @@ class LayoutPopup(KeyboardPopup, LayoutView, TouchInput):
         else:
             self.close_window()
 
-    def _on_draw(self, widget, context):
-        decorated = LayoutView.draw(self, widget, context)
+    def on_draw(self, widget, context):
+        context.push_group()
+
+        LayoutView.draw(self, widget, context)
+
+        context.pop_group_to_source()
+        context.paint_with_alpha(self._opacity)
 
     def draw_window_frame(self, context, lod):
         corner_radius = config.CORNER_RADIUS
@@ -455,9 +524,10 @@ class LayoutBuilderKeySequence(LayoutBuilder):
     @staticmethod
     def build(source_key, color_scheme, key_sequence):
         # parse sequence into lines
-        lines, ncolumns = LayoutBuilderKeySequence \
-                             ._layout_sequence(key_sequence)
-        return LayoutBuilderKeySequence._create_layout(source_key, color_scheme,
+        lines, ncolumns = \
+            LayoutBuilderKeySequence ._layout_sequence(key_sequence)
+        return LayoutBuilderKeySequence._create_layout(source_key,
+                                                       color_scheme,
                                                        lines, ncolumns)
 
     @staticmethod
@@ -474,7 +544,7 @@ class LayoutBuilderKeySequence(LayoutBuilder):
                               rect.w * ncolumns + spacing[0] * (ncolumns - 1),
                               rect.h * nrows + spacing[1] * (nrows - 1))
 
-        # subdive into logical rectangles for the keys
+        # subdivide into logical rectangles for the keys
         layout_rect = context.canvas_to_log_rect(layout_canvas_rect)
         key_rects = layout_rect.subdivide(ncolumns, nrows, *spacing)
 
@@ -583,3 +653,67 @@ class LayoutBuilderAlternatives(LayoutBuilderKeySequence):
         return LayoutBuilderKeySequence.build(source_key, color_scheme,
                                               key_sequence)
 
+
+class PendingSeparatorPopup(KeyboardPopup):
+    """ Ephemeral popup displaying the pending word separator. """
+
+    _osk_util = osk.Util()
+
+    def __init__(self):
+        KeyboardPopup.__init__(self)
+        self.connect("realize", self._on_realize_event)
+        self._visible = False
+
+    def show_at(self, view, character_rect):
+        toplevel = view.get_toplevel()
+
+        self.set_transient_for(toplevel)
+        self.realize()
+
+        x, y, w, h = character_rect
+        self.set_default_size(w, h)
+        self.resize(w, h)
+        self.move(x, y)
+
+        self.supports_alpha = view.supports_alpha
+        if self.supports_alpha:
+            self.set_opacity(toplevel.get_opacity())
+        self.show_all()
+        self._visible = True
+
+    def hide(self):
+        KeyboardPopup.hide(self)
+        self._visible = False
+
+    def is_visible(self):
+        return self._visible
+
+    def _on_realize_event(self, user_data):
+        self.set_override_redirect(True)
+
+        # set minimal input shape for the popup to become click-through
+        win = self.get_window()
+        self._osk_util.set_input_rect(win, 0, 0, 1, 1)
+
+    def on_draw(self, widget, cr):
+        rect = Rect(0, 0,
+                    self.get_allocated_width(), self.get_allocated_height())
+
+        fill_rgba = (1.0, 0.5, 0.5, 0.2)
+        stroke_rgba = (1.0, 0.5, 0.5, 0.7)
+
+        cr.set_source_rgba(*fill_rgba)
+        cr.rectangle(*rect)
+        cr.fill()
+
+        r = rect
+        top = r.y + r.h * 0.75
+        bottom = r.bottom()
+        cr.set_source_rgba(*stroke_rgba)
+        cr.set_line_width(max(1, r.h / 6))
+        cr.move_to(r.left(), top)
+        cr.line_to(r.left(), bottom)
+        cr.line_to(r.right(), bottom)
+        cr.line_to(r.right(), top)
+        cr.stroke()
+
