@@ -4,7 +4,7 @@
 # Copyright © 2007-2010 Chris Jones <tortoise@tortuga>
 # Copyright © 2008-2010 Francesco Fumanti <francesco.fumanti@gmx.net>
 # Copyright © 2012-2013 Gerd Kohlberger <lowfi@chello.at>
-# Copyright © 2010-2015 marmuta <marmvta@gmail.com>
+# Copyright © 2010-2016 marmuta <marmvta@gmail.com>
 #
 # This file is part of Onboard.
 #
@@ -39,6 +39,7 @@ try:
     import dbus
     import dbus.service
     import dbus.mainloop.glib
+    from Onboard.DBusUtils import ServiceBase, dbus_property
 except ImportError:
     pass
 
@@ -71,8 +72,6 @@ class OnboardGtk(object):
     Main controller class for Onboard using GTK+
     """
 
-    DBUS_NAME = "org.onboard.Onboard"
-
     keyboard = None
 
     def __init__(self):
@@ -92,8 +91,8 @@ class OnboardGtk(object):
             # Use D-bus main loop by default
             dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)
 
-            # Don't fail to start in Xenial's lightdm due
-            # to D-Bus not available.
+            # Don't fail to start in Xenial's lightdm when
+            # D-Bus isn't available.
             try:
                 bus = dbus.SessionBus()
             except dbus.exceptions.DBusException:
@@ -114,7 +113,8 @@ class OnboardGtk(object):
 
         if bus:
             # Check if there is already an Onboard instance running
-            has_remote_instance = bus.name_has_owner(self.DBUS_NAME)
+            has_remote_instance = \
+                bus.name_has_owner(ServiceOnboardKeyboard.NAME)
 
             # Onboard in Ubuntu on first start silently embeds itself into
             # gnome-screensaver and stays like this until embedding is manually
@@ -134,14 +134,15 @@ class OnboardGtk(object):
                 if has_remote_instance and \
                    not config.options.allow_multiple_instances:
                     # Present remote instance
-                    remote = bus.get_object(self.DBUS_NAME,
+                    remote = bus.get_object(ServiceOnboardKeyboard.NAME,
                                             ServiceOnboardKeyboard.PATH)
                     remote.Show(dbus_interface=ServiceOnboardKeyboard.IFACE)
                     _logger.info("Exiting: Not the primary instance.")
                     sys.exit(0)
 
                 # Register our dbus name
-                self._bus_name = dbus.service.BusName(self.DBUS_NAME, bus)
+                self._bus_name = \
+                    dbus.service.BusName(ServiceOnboardKeyboard.NAME, bus)
 
         self.init()
 
@@ -163,6 +164,7 @@ class OnboardGtk(object):
 
     def init(self):
         self.keyboard_state = None
+        self._last_mod_mask = 0
         self.vk_timer = None
         self.reset_vk()
         self._connections = []
@@ -174,6 +176,19 @@ class OnboardGtk(object):
         # finish config initialization
         config.init()
 
+        # Optionally wait a little before proceeding.
+        # When the system starts up, the docking strut can be unreliable
+        # on Compiz (Yakkety) and the keyboard may end up at unexpected initial
+        # positions. Delaying the startup when launched by
+        # onboard-autostart.desktop helps to prevent this.
+        delay = config.startup_delay
+        if delay:
+            Timer(delay, self._init_delayed)
+        else:
+            self._init_delayed()
+
+    def _init_delayed(self):
+
         # Release pressed keys when onboard is killed.
         # Don't keep enter key stuck when being killed by lightdm.
         self._osk_util = osk.Util()
@@ -299,6 +314,9 @@ class OnboardGtk(object):
         config.window.enable_inactive_transparency_notify_add(update_transparency)
         config.window.inactive_transparency_notify_add(update_inactive_transparency)
         config.window.docking_notify_add(self._update_docking)
+        config.window.docking_aspect_change_range_notify_add(
+            lambda x: self.keyboard_widget
+            .update_docking_aspect_change_range())
 
         # layout
         config.layout_filename_notify_add(reload_layout)
@@ -526,7 +544,15 @@ class OnboardGtk(object):
         mod_mask = keymap.get_modifier_state()
         _logger.debug("keyboard state changed to 0x{:x}" \
                       .format(mod_mask))
-        self.keyboard.set_modifiers(mod_mask)
+        if self._last_mod_mask == mod_mask:
+            # Must be something other than changed modifiers,
+            # group change on wayland, for example.
+            self.reload_layout_delayed()
+        else:
+            self.keyboard.set_modifiers(mod_mask)
+
+        self._last_mod_mask = mod_mask
+
 
     def cb_vk_timer(self):
         """
@@ -654,7 +680,7 @@ class OnboardGtk(object):
         if vk:
             try:
                 vk.reload() # reload keyboard names
-                keyboard_state = (vk.get_layout_symbols(),
+                keyboard_state = (vk.get_layout_as_string(),
                                   vk.get_current_group_name())
             except osk.error:
                 self.reset_vk()
@@ -795,100 +821,51 @@ class OnboardGtk(object):
 
 
 if "dbus" in globals():
-    class ServiceOnboardKeyboard(dbus.service.Object):
+    class ServiceOnboardKeyboard(ServiceBase):
         """
-        Onboard's D-Bus service.
+        Onboard's main D-Bus service.
         """
 
+        NAME = "org.onboard.Onboard"
         PATH = "/org/onboard/Onboard/Keyboard"
         IFACE = "org.onboard.Onboard.Keyboard"
 
-        class ServiceOnboardException(dbus.DBusException):
-            _dbus_error_name = 'org.onboard.Exception'
-
-
         def __init__(self, app):
-            super(ServiceOnboardKeyboard, self).__init__(dbus.SessionBus(), self.PATH)
+            ServiceBase.__init__(self, dbus.SessionBus(), self.NAME, self.PATH)
             self._keyboard = app.keyboard
             self._keyboard_widget = app.keyboard_widget
 
         @dbus.service.method(dbus_interface=IFACE)
-        def Show(self):
+        def Show(self):  # noqa: flake8
             self._keyboard.set_visible(True)
 
         @dbus.service.method(dbus_interface=IFACE)
-        def Hide(self):
+        def Hide(self):  # noqa: flake8
             self._keyboard.set_visible(False)
 
         @dbus.service.method(dbus_interface=IFACE)
-        def ToggleVisible(self):
+        def ToggleVisible(self):  # noqa: flake8
             self._keyboard.toggle_visible()
 
         # private method, for unit-testing only
         if config.is_running_from_source:
             @dbus.service.method(dbus_interface=IFACE,
-                                 out_signature='a(sada{is}a{sb})')
+                                 out_signature='a(sada{is}a{sb})')  # noqa: flake8
             def GetKeyState(self):
                 result = []
                 for key in self._keyboard.iter_keys():
                     r = self._keyboard_widget.get_key_screen_rect(key)
                     if not r.is_empty() and key.is_path_visible():
-                        labels = {m:l for m, l in key.labels.items() if l}  # -None
+                        labels = {m : l for m, l in key.labels.items() if l}
                         result.append([key.get_id(),
-                                    list(r),
-                                    labels,
-                                    key.get_state()])
+                                       list(r),
+                                       labels,
+                                       key.get_state()])
                 return result
 
-        @dbus.service.method(dbus_interface=dbus.PROPERTIES_IFACE,
-                             in_signature='ss', out_signature='v')
-        def Get(self, iface, prop):
-            if iface == self.IFACE:
-                if prop == 'Visible':
-                    return self._keyboard.is_visible()
-                else:
-                    raise self.ServiceOnboardException(\
-                        ('Unknown property \'{0}\'').format(prop))
-            else:
-                raise self.ServiceOnboardException(\
-                    ('Unknown interface \'{0}\'').format(iface))
-
-        @dbus.service.method(dbus_interface=dbus.PROPERTIES_IFACE, in_signature='ssv')
-        def Set(self, iface, prop, value):
-            if iface == self.IFACE:
-                if prop == 'Visible':
-                    raise self.ServiceOnboardException(\
-                        ('Property \'{0}\' is read-only').format(prop))
-                else:
-                    raise self.ServiceOnboardException(\
-                        ('Unknown property \'{0}\'').format(prop))
-            else:
-                raise self.ServiceOnboardException(\
-                    ('Unknown interface \'{0}\'').format(iface))
-
-        @dbus.service.method(dbus_interface=dbus.PROPERTIES_IFACE,
-                             in_signature='s', out_signature='a{sv}')
-        def GetAll(self, iface):
-            if iface == self.IFACE:
-                return { 'Visible': self._keyboard.is_visible() }
-            else:
-                raise self.ServiceOnboardException(\
-                    ('Unknown interface \'{0}\'').format(iface))
-
-        @dbus.service.method(dbus_interface=dbus.INTROSPECTABLE_IFACE, out_signature='s')
-        def Introspect(self):
-            ref = dbus.service.Object.Introspect(self, self._object_path, self.connection)
-
-            iface = '  <interface name="{}">\n' \
-                    '      <property name="Visible" type="b" access="read"/>\n' \
-                    '  </interface>\n' \
-                    .format(self.IFACE)
-
-            return ref[:-8] + iface + '</node>\n'
-
-        @dbus.service.signal(dbus_interface=dbus.PROPERTIES_IFACE, signature='sa{sv}as')
-        def PropertiesChanged(self, iface, changed, invalidated):
-            return iface, changed, invalidated
+        @dbus_property(dbus_interface=IFACE, signature='b')
+        def Visible(self):  # noqa: flake8
+            return self._keyboard.is_visible()
 
 
 def cb_any_event(event, onboard):
@@ -900,7 +877,7 @@ def cb_any_event(event, onboard):
     except ValueError:
         type = None
 
-    if 0: # debug
+    if 0:  # debug
         a = [event, event.type]
         if type == Gdk.EventType.VISIBILITY_NOTIFY:
             a += [event.state]
